@@ -2,11 +2,18 @@ package Gtk2::CV;
 
 use XSLoader;
 
-$VERSION = '0.3';
+$VERSION = '0.5';
 
 XSLoader::load "Gtk2::CV", $VERSION;
 
 use Gtk2;
+use Glib;
+
+use IO::AIO;
+
+add_watch Glib::IO IO::AIO::poll_fileno,
+                   in => sub { IO::AIO::poll_cb, 1 };
+IO::AIO::max_outstanding 128;
 
 sub find_rcfile($) {
    my $path;
